/*---------------------------------------------------------------------------------
 bbEdgeHook ( 2004 Slade Taylor [bladestaylor@yahoo.com])
 ----------------------------------------------------------------------------------
 bbEdgeHook is a plugin for Blackbox for Windows.  For more information,
 please visit [http://bb4win.org] or [http://sourceforge.net/projects/bb4win].
 ----------------------------------------------------------------------------------
 bbEdgeHook is free software, released under the GNU General Public License,
 version 2, as published by the Free Software Foundation.  It is distributed
 WITHOUT ANY WARRANTY--without even the implied warranty of MERCHANTABILITY
 or FITNESS FOR A PARTICULAR PURPOSE.  Please see the GNU General Public
 License for more details:  [http://www.fsf.org/licenses/gpl.html].
 --------------------------------------------------------------------------------*/

#include "bbEdgeHook.h"
#include "xutil/SystemMenuItem.h"

//---------------------------------------------------------------------------------

bool WINAPI DllMain(HINSTANCE hinstance, DWORD fdwReason, LPVOID lpvReserved)
{
    if (fdwReason == DLL_PROCESS_ATTACH)
        DisableThreadLibraryCalls(hInstance = hinstance);
    return true;
}

//---------------------------------------------------------------------------------

int beginPlugin(HINSTANCE hPluginInstance)
{
    // --- load xutil.dll -------------------------------
    char xutil_path[MAX_LINE_LENGTH];
    strcpy(xutil_path + GetModuleFileName(hPluginInstance, xutil_path, MAX_LINE_LENGTH) - 14, "xutil.dll");
    if (!xutil_Initialize(xutil_path))
    {
        MessageBox(NULL, "An error occurred while loading xutil.dll.", pluginInfo(APPNAME), MB_ICONERROR);
        return 1;
    }
    // --------------------------------------------------

    WNDCLASS wc;
    ZeroMemory(&wc, sizeof(wc));
    wc.lpfnWndProc = PluginProc;
    wc.hInstance = hPluginInstance;
    wc.lpszClassName = pluginInfo(APPNAME);
    wc.hCursor = NULL;
    RegisterClass(&wc);

    hPluginWnd = CreateWindowEx(  // create window
        WS_EX_TOOLWINDOW,
        pluginInfo(APPNAME),
        NULL,
        WS_POPUP,
        0, 0, 0, 0,
        NULL,
        NULL,
        hPluginInstance,
        NULL);

    xutil_RC();
    bWin2k = IsWin2k();
    ManageDTHook();

    BBThread = GetCurrentThreadId();  // get blackbox thread
    hMouseHook = SetWindowsHookEx(WH_MOUSE, MouseProc, hPluginInstance, 0);  // set hook
    SendMessage(hCoreWnd, BB_REGISTERMESSAGE, (WPARAM)hPluginWnd, (LPARAM)bb_messages);
    return 0;
}

//---------------------------------------------------------------------------------

void endPlugin(HINSTANCE hPluginInstance)
{
    DestroyWindow(hPluginWnd);              // destroy window
    SendMessage(hCoreWnd, BB_UNREGISTERMESSAGE, (WPARAM)hPluginWnd, (LPARAM)bb_messages);
    UnregisterClass(pluginInfo(APPNAME), hPluginInstance);
    if (hMouseHook)
        UnhookWindowsHookEx(hMouseHook);    // remove hook
    xutil_Unload();
}

//---------------------------------------------------------------------------------

LRESULT CALLBACK PluginProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message)
    {
        //---------------------------

        case BB_BROADCAST:
            ProcessBroam((char*)lParam);
            break;

        //---------------------------

        case BB_RECONFIGURE:
            xutil_RC();
            break;

        //---------------------------

        default:
            return DefWindowProc(hwnd, message, wParam, lParam);
    }
    return 0;
}

//---------------------------------------------------------------------------------

void ProcessBroam(char *broam_txt)
{
    if (_strnicmp(broam_txt, "@bbEdgeHook.", 12))
        return;
    broam_txt += 12;
    if (!_stricmp(broam_txt, "Pause"))
        goto br_pause;
    if (!_stricmp(broam_txt, "Resume"))
        goto br_resume;
    if (!_stricmp(broam_txt, "Toggle"))
    {
        if (hMouseHook)
        {
br_pause:   UnhookWindowsHookEx(hMouseHook);  // remove hook
            hMouseHook = NULL;
        }
        else
        {
br_resume:  hMouseHook = SetWindowsHookEx(WH_MOUSE, MouseProc, hInstance, 0);  // set hook
            bInTitlebar = false;
        }
    }
}

//---------------------------------------------------------------------------------

bool ProcessMousewheel(POINT pt, bool direction)
{
    if (EnableMarginMouseWheel && !PtInRect(&ExclusionRect, pt))
    {
        // don't scroll here if DesktopHook.dll has already done it
        if (!bDTHook || (WindowFromPoint(pt) != GetDesktopWindow()))
        {
            PostMessage(  // scroll through workspaces
                hCoreWnd,
                BB_WORKSPACE,
                direction ? Direction_Desk.y: Direction_Desk.x,
                NULL);
            return true;
        }
    }
    if (EnableTitlebarMouseWheel && bInTitlebar)
    {
        const HWND h = WindowFromPoint(pt);
        if (h && (GetWindowThreadProcessId(h, NULL) != BBThread))
        {
            ForceForegroundWindow(hCoreWnd);
            ForceForegroundWindow(h);
            PostMessage(  // move window to new workspace
                hCoreWnd,
                BB_WORKSPACE,
                direction ? Direction_Window.y: Direction_Window.x,
                (LPARAM)h);
            return true;
        }
    }
    return false;
}

//---------------------------------------------------------------------------------

LRESULT CALLBACK MouseProc(int nCode, WPARAM wParam, LPARAM lParam)
{
    if (nCode >= 0) switch (wParam)
    {
        //---------------------------

        case WM_MOUSEWHEEL:
            if (bWin2k)
            {
                MOUSEHOOKSTRUCTEX *pData = (MOUSEHOOKSTRUCTEX*)lParam;
                if (ProcessMousewheel(pData->pt, ((short)HIWORD(pData->mouseData) < 0)))
                    return 1;
            }
            else
            {
                MOUSEHOOKSTRUCT *pData = (MOUSEHOOKSTRUCT*)lParam;
                if (ProcessMousewheel(pData->pt, ((short)LOWORD(pData->dwExtraInfo) < 0)))
                    return 1;
            }
            break;

        //---------------------------

        case WM_MOUSEMOVE:
            bInTitlebar = false;
            break;

        //---------------------------

        case WM_NCMOUSEMOVE:
            bInTitlebar = CheckHittest(((MOUSEHOOKSTRUCT*)lParam)->wHitTestCode);
            break;

        //---------------------------

        case WM_RBUTTONDOWN:
            if (IsValid((MOUSEHOOKSTRUCT*)lParam))
                return 1;
            break;

        //---------------------------

        case WM_NCRBUTTONDOWN:
            {
                const MOUSEHOOKSTRUCT *pData = (MOUSEHOOKSTRUCT*)lParam;
                if (IsValid(pData))
                    return 1;

                if ((EnableGestures || EnableSystemMenu) && CheckHittest(pData->wHitTestCode))
                {
                    GestureStart = pData->pt.x;
                    return 1;
                }
            }
            break;

        //---------------------------

        case WM_RBUTTONUP:
            {
                GestureStart = -1;
                MOUSEHOOKSTRUCT *pData = (MOUSEHOOKSTRUCT*)lParam;
                if (IsValid(pData))
                {
                    PostMessage(hCoreWnd, BB_MENU, 0, 0);  // show desktop menu
                    return 1;
                }
                if (((GetAsyncKeyState(VK_SHIFT) & 0x8000) != 0) &&  // if holding down ctrl and shift
                    ((GetAsyncKeyState(VK_CONTROL) & 0x8000) != 0))
                {
                    if (EnableSystemMenu && (GetWindowThreadProcessId(pData->hwnd, NULL) != BBThread))
                    {
                        SetForegroundWindow(pData->hwnd);
                        PostMessage(
                            FindWindow("SystemMenuItem", NULL),
                            SystemMenuItem::MSG_SYSTEM_MENU_SHOW,
                            NULL,
                            (LPARAM)pData->hwnd);
                        return 1;
                    }
                }
            }
            break;

        //---------------------------

        case WM_NCRBUTTONUP:
            {
                const MOUSEHOOKSTRUCT *pData = (MOUSEHOOKSTRUCT*)lParam;
                if (IsValid(pData))
                {
                    PostMessage(hCoreWnd, BB_MENU, 0, 0);  // show desktop menu
                    return 1;
                }
                if (CheckHittest(pData->wHitTestCode))
                {
                    if (EnableGestures)
                    {
                        if (GestureStart > 0)
                        {
                            const int thres = 10;
                            WPARAM wp = BBWS_MOVEWINDOWLEFT;
                            if (pData->pt.x < (GestureStart - thres))
                                goto p_gesture;
                            if (pData->pt.x > (GestureStart + thres))
                            {
                                wp = BBWS_MOVEWINDOWRIGHT;
p_gesture:                      PostMessage(hCoreWnd, BB_WORKSPACE, wp, (LPARAM)pData->hwnd);
                                GestureStart = -1;
                                return 1;
                            }
                        }
                        GestureStart = -1;
                    }
                    if (EnableSystemMenu && (GetWindowThreadProcessId(pData->hwnd, NULL) != BBThread))
                    {
                        SetForegroundWindow(pData->hwnd);
                        PostMessage(
                            FindWindow("SystemMenuItem", NULL),
                            SystemMenuItem::MSG_SYSTEM_MENU_SHOW,
                            NULL,
                            (LPARAM)pData->hwnd);
                        return 1;
                    }
                }
            }
            break;

        //---------------------------

        case WM_NCMBUTTONUP:
            {
                const MOUSEHOOKSTRUCT *pData = (MOUSEHOOKSTRUCT*)lParam;
                if (CheckHittest(pData->wHitTestCode))
                {
                    PostMessage(hCoreWnd, BB_WINDOWCLOSE, 0, (LPARAM)pData->hwnd);
                    return 1;
                }
            }

        //---------------------------
    }
    return CallNextHookEx(hMouseHook, nCode, wParam, lParam);
}

//---------------------------------------------------------------------------------

bool IsWin2k()
{
    OSVERSIONINFO osInfo;
    ZeroMemory(&osInfo, sizeof(osInfo));
    osInfo.dwOSVersionInfoSize = sizeof(osInfo);
    GetVersionEx(&osInfo);
    return ((osInfo.dwPlatformId == VER_PLATFORM_WIN32_NT) && (osInfo.dwMajorVersion >= 5));
}

//---------------------------------------------------------------------------------

bool IsValid(const MOUSEHOOKSTRUCT *pData)
{
    return !PtInRect(&ExclusionRect, pData->pt) &&
        (GetWindowThreadProcessId(pData->hwnd, NULL) != BBThread);
}

//---------------------------------------------------------------------------------

void ForceForegroundWindow(HWND hwnd)
{
    HWND fw = GetForegroundWindow();
    if (hwnd == fw)
        return;

    DWORD   id_1 = GetWindowThreadProcessId(fw, NULL),
            id_2 = GetCurrentThreadId();

    if (id_1 == id_2)
        SetForegroundWindow(hwnd);
    else
    {
        AttachThreadInput(id_1, id_2, true);
        SetForegroundWindow(hwnd);
        AttachThreadInput(id_1, id_2, false);
    }
}

//---------------------------------------------------------------------------------

void GetScreenSettings()
{
    int v_screen_x = 0, v_screen_y = 0;
    const int   screen_width = GetSystemMetrics(SM_CXSCREEN),
                screen_height = GetSystemMetrics(SM_CYSCREEN);
    int         v_screen_width = GetSystemMetrics(SM_CXVIRTUALSCREEN),
                v_screen_height = GetSystemMetrics(SM_CYVIRTUALSCREEN);

    if ((v_screen_width > screen_width) || (v_screen_height > screen_height))
    {
        v_screen_x = GetSystemMetrics(SM_XVIRTUALSCREEN);
        v_screen_y = GetSystemMetrics(SM_YVIRTUALSCREEN);
    }
    else  // if not multimon
    {
        v_screen_width  = screen_width;
        v_screen_height = screen_height;
    }

    ExclusionRect.top = v_screen_y + MarginRect.top;
    ExclusionRect.bottom = v_screen_y + v_screen_height - MarginRect.bottom;
    ExclusionRect.left = v_screen_x + MarginRect.left;
    ExclusionRect.right = v_screen_x + v_screen_width - MarginRect.right - 1;
}

//---------------------------------------------------------------------------------

bool CheckHittest(UINT ht)
{
    switch (ht)
    {
        //--------------------
        //case HTERROR:         // (-2)
        //case HTTRANSPARENT:   // (-1)
        //case HTNOWHERE:       // 0
        //case HTCLIENT:        // 1
        case HTCAPTION:         // 2
        case HTSYSMENU:         // 3
        //case HTGROWBOX:       // 4
        //case HTSIZE:          // HTGROWBOX
        //case HTMENU:          // 5
        //case HTHSCROLL:       // 6
        //case HTVSCROLL:       // 7
        case HTMINBUTTON:       // 8
        case HTMAXBUTTON:       // 9
        case HTLEFT:            // 10
        case HTRIGHT:           // 11
        case HTTOP:             // 12
        case HTTOPLEFT:         // 13
        case HTTOPRIGHT:        // 14
        case HTBOTTOM:          // 15
        case HTBOTTOMLEFT:      // 16
        case HTBOTTOMRIGHT:     // 17
        //case HTBORDER:        // 18
        //case HTREDUCE:        // HTMINBUTTON
        //case HTZOOM:          // HTMAXBUTTON
        //case HTSIZEFIRST:     // HTLEFT
        //case HTSIZELAST:      // HTBOTTOMRIGHT
        //case HTOBJECT:        // 19  //#if(WINVER >= 0x0400)
        case HTCLOSE:           // 20  //#if(WINVER >= 0x0400)
        //case HTHELP:          // 21  //#if(WINVER >= 0x0400)
        //--------------------
        case 100:               // bbWinSkin HTROLLUP
        case 101:               // bbWinSkin HTONTOP
        case 102:               // bbWinSkin HTICON
        //--------------------
        case 103:               // bbLeanSkin button
        case 104:               // bbLeanSkin button
        case 105:               // bbLeanSkin button
        //--------------------
            return true;
    }
    return false;
}

//---------------------------------------------------------------------------------

void ManageDTHook()
{
    if (hCoreWnd = FindWindow("BlackBoxClass", "BlackBox"))
    {
        // if the core is 0.0.90 and the OS is win2k or later
        if (!FindWindow("BBLeanDesktop", NULL) && bWin2k)
            // Don't process mousewheel in the margin over the desktop
            // because DesktopHook.dll with have already switched workspaces
            bDTHook = true;
    }
    else
        hCoreWnd = FindWindow("xoblite", NULL);
}

//---------------------------------------------------------------------------------

bool xutil_Initialize(char *path)
{
    if (xutil_hmodule = LoadLibrary(path))
    {
        *(FARPROC*)&xutil_pRC_Bool  = GetProcAddress(xutil_hmodule, "RC_Bool");
        *(FARPROC*)&xutil_pRC_Int   = GetProcAddress(xutil_hmodule, "RC_Int");
        *(FARPROC*)&xutil_pStart    = GetProcAddress(xutil_hmodule, "Start");
        *(FARPROC*)&xutil_pEnd      = GetProcAddress(xutil_hmodule, "End");
        if (xutil_pEnd && xutil_pStart && xutil_pRC_Int && xutil_pRC_Bool)
        {
            xutil_pStart();
            return true;
        }
    }
    return false;
};

//---------------------------------------------------------------------------------

void xutil_Unload()
{
    if (xutil_pEnd)     xutil_pEnd();
    if (xutil_hmodule)  FreeLibrary(xutil_hmodule);
};

//---------------------------------------------------------------------------------

void xutil_RC()
{
    if (!xutil_pRC_Int || !xutil_pRC_Bool)
        return;

    char rc_path[MAX_LINE_LENGTH];
    const int i = GetModuleFileName(hInstance, rc_path, MAX_LINE_LENGTH);
    if (i > 3) { rc_path[i - 3] = 'r'; rc_path[i - 2] = 'c'; rc_path[i - 1] = 0; }

    xutil_pRC_Int(rc_path,  "Margin.Left:",                 &MarginRect.left,           1);
    xutil_pRC_Int(rc_path,  "Margin.Right:",                &MarginRect.right,          1);
    xutil_pRC_Int(rc_path,  "Margin.Top:",                  &MarginRect.top,            0);
    xutil_pRC_Int(rc_path,  "Margin.Bottom:",               &MarginRect.bottom,         0);
    xutil_pRC_Bool(rc_path, "Titlebar.Menu.Enable:",        &EnableSystemMenu,          true);
    xutil_pRC_Bool(rc_path, "Titlebar.Gestures.Enable:",    &EnableGestures,            true);
    xutil_pRC_Bool(rc_path, "MouseWheel.Margin.Enable:",    &EnableMarginMouseWheel,    true);
    xutil_pRC_Bool(rc_path, "MouseWheel.Titlebar.Enable:",  &EnableTitlebarMouseWheel,  true);
    xutil_pRC_Bool(rc_path, "MouseWheel.ReverseDirection:", &ReverseDirection,          false);

    if (ReverseDirection)
    {
        Direction_Desk.x = BBWS_DESKRIGHT,          Direction_Desk.y = BBWS_DESKLEFT;
        Direction_Window.x = BBWS_MOVEWINDOWRIGHT,  Direction_Window.y = BBWS_MOVEWINDOWLEFT;
    }
    else
    {
        Direction_Desk.x = BBWS_DESKLEFT,           Direction_Desk.y = BBWS_DESKRIGHT;
        Direction_Window.x = BBWS_MOVEWINDOWLEFT,   Direction_Window.y = BBWS_MOVEWINDOWRIGHT;
    }

    GetScreenSettings();
}

//---------------------------------------------------------------------------------
