/*---------------------------------------------------------------------------------
 ID_Item ( 2004 Slade Taylor [bladestaylor@yahoo.com])
 ----------------------------------------------------------------------------------
 ID_Item is a class for Blackbox for Windows.  For more information,
 please visit [http://bb4win.org] or [http://sourceforge.net/projects/bb4win].
 ----------------------------------------------------------------------------------
 ID_Item is free software, released under the GNU General Public License,
 version 2, as published by the Free Software Foundation.  It is distributed
 WITHOUT ANY WARRANTY--without even the implied warranty of MERCHANTABILITY
 or FITNESS FOR A PARTICULAR PURPOSE.  Please see the GNU General Public
 License for more details:  [http://www.fsf.org/licenses/gpl.html].
 --------------------------------------------------------------------------------*/

#include "stdafx.h"
#include "ID_Item.h"

//---------------------------------------------------------------------------------

ID_Item*    ID_Item::m_p_head                                   = NULL;
bool        ID_Item::m_idEnabled                                = true;
int         ID_Item::m_update_id_counter                        = 0;
int         ID_Item::m_unique_id_counter                        = 0;
int*        ID_Item::m_p_current_counter                        = NULL;
tasklist*   (*ID_Item::m_pGetTaskListPtr)()                     = NULL;
Menu*       (*ID_Item::m_pMakeNamedMenu)(LPCSTR, LPCSTR, bool)  = NULL;

//---------------------------------------------------------------------------------

ID_Item::ID_Item(HWND hwnd)
{
    m_hwnd = hwnd;
    m_id = *m_p_current_counter - 1;
    m_next = NULL;
}

//---------------------------------------------------------------------------------

ID_Item::~ID_Item()
{
    if (m_next)
        delete m_next;
}

//---------------------------------------------------------------------------------

ID_Item *ID_Item::ListStartup()
{
    *(FARPROC*)&m_pMakeNamedMenu = GetProcAddress(GetModuleHandle("blackbox.exe"), "MakeNamedMenu");
    *(FARPROC*)&m_pGetTaskListPtr = GetProcAddress(GetModuleHandle("blackbox.exe"), "GetTaskListPtr");
    m_idEnabled = (m_pMakeNamedMenu != NULL) && (m_pGetTaskListPtr != NULL);
    m_p_current_counter = &m_unique_id_counter;
    return m_p_head = new ID_Item(NULL);
}

//---------------------------------------------------------------------------------

void ID_Item::ListShutdown()
{
    if (m_p_head)
        delete m_p_head;
}

//---------------------------------------------------------------------------------

void ID_Item::PushBack(HWND hwnd)
{
    if (m_idEnabled && (m_p_current_counter == &m_unique_id_counter))
    {
        ID_Item *tmp = this;
        while (tmp->m_next) tmp = tmp->m_next;
        tmp->m_next = new ID_Item(hwnd);
    }
};

//---------------------------------------------------------------------------------

ID_Item* ID_Item::Erase()
{
    ID_Item *tmp = m_next->m_next;
    m_next->m_next = NULL;
    delete m_next;
    return m_next = tmp;
};

//---------------------------------------------------------------------------------

void ID_Item::CleanList()
{
    if (!m_idEnabled)
        return;

    tasklist *t = NULL;
    ID_Item *p = this;
    while (p->m_next)  // clean out old list-objects; find id of menu to update
    {
        if (t = m_pGetTaskListPtr())
            do if (t->hwnd == p->m_next->m_hwnd)
            {
                p = p->m_next;
                goto o_ct;
            }
            while (t = t->next);
        p->Erase();
o_ct:;
    }
}

//---------------------------------------------------------------------------------

Menu* ID_Item::XMakeMenu(LPSTR header, bool popup)
{
    if (!m_idEnabled)
        return MakeMenu(header);
    char buf[50];
    sprintf(buf, "BBEH_SM_%d", *m_p_current_counter);
    ++*m_p_current_counter;
    return m_pMakeNamedMenu(header, buf, popup);
}

//---------------------------------------------------------------------------------

void ID_Item::XDelMenu(Menu *menu)
{
    if (!m_idEnabled && menu)
        DelMenu(menu);
}

//---------------------------------------------------------------------------------

bool ID_Item::SetUpdateID(HWND hwnd, int id)
{
    if (!m_idEnabled)
        return false;

    ID_Item *p = m_next;
    while (p)
    {
        if (p->m_id == id)
        {
            m_update_id_counter = id;
            return true;
        }
        p = p->m_next;
    }
    return false;
}

//---------------------------------------------------------------------------------

int ID_Item::SetCurrentIDCounter(bool popup)
{
    m_p_current_counter = &(popup ? m_unique_id_counter: m_update_id_counter);
    return *m_p_current_counter;
}

//---------------------------------------------------------------------------------
