/*---------------------------------------------------------------------------------
 SystemMenuItem ( 2004 Slade Taylor [bladestaylor@yahoo.com])
 ----------------------------------------------------------------------------------
 SystemMenuItem is a class for Blackbox for Windows.  For more information,
 please visit [http://bb4win.org] or [http://sourceforge.net/projects/bb4win].
 ----------------------------------------------------------------------------------
 SystemMenuItem is free software, released under the GNU General Public License,
 version 2, as published by the Free Software Foundation.  It is distributed
 WITHOUT ANY WARRANTY--without even the implied warranty of MERCHANTABILITY
 or FITNESS FOR A PARTICULAR PURPOSE.  Please see the GNU General Public
 License for more details:  [http://www.fsf.org/licenses/gpl.html].
 --------------------------------------------------------------------------------*/

#include "stdafx.h"
#include <stdlib.h>
#include "SystemMenuItem.h"

//--------------------------------------

#define szClassName "SystemMenuItem"

//--------------------------------------

static SystemMenuItem *_pSystemMenu = NULL;                             // store pointer for wndproc
static int bb_messages[] = {BB_BROADCAST, BB_RECONFIGURE, 0};           // msgs to register

//--------------------------------------

static LRESULT CALLBACK _SystemMenuProc(HWND, UINT, WPARAM, LPARAM);    // declaration for wndproc

//--------------------------------------

void*   (*SystemMenuItem::m_pGetSettingPtr)(int)            = NULL;
HFONT   (*SystemMenuItem::m_pCreateStyleFont)(StyleItem*)   = NULL;

//---------------------------------------------------------------------------------

void SystemMenuItem::Startup()
{
    *(FARPROC*)&m_pGetSettingPtr = GetProcAddress(GetModuleHandle("blackbox.exe"), "GetSettingPtr");
    *(FARPROC*)&m_pCreateStyleFont = GetProcAddress(GetModuleHandle("blackbox.exe"), "CreateStyleFont");
    if (m_pGetSettingPtr && m_pCreateStyleFont)
        _pSystemMenu = new SystemMenuItem();
}

//---------------------------------------------------------------------------------

void SystemMenuItem::Shutdown()
{
    if (_pSystemMenu)
        delete _pSystemMenu;
}

//---------------------------------------------------------------------------------

SystemMenuItem::SystemMenuItem()
{
    m_bbMenu = NULL;
    m_hinstance = GetModuleHandle(NULL);
    m_bbThread = GetCurrentThreadId();

    //---------------------

    WNDCLASS wc;
    ZeroMemory(&wc, sizeof(wc));
    wc.lpfnWndProc = _SystemMenuProc;
    wc.hInstance = m_hinstance;
    wc.lpszClassName = szClassName;
    wc.hCursor = NULL;
    RegisterClass(&wc);

    m_hwnd = CreateWindowEx(  // create window
        WS_EX_TOOLWINDOW,
        szClassName,
        NULL,
        WS_POPUP,
        0, 0, 0, 0,
        NULL,
        NULL,
        m_hinstance,
        NULL);

    //---------------------

    ZeroMemory((void*)&m_menuInfo, sizeof(m_menuInfo));
    m_menuInfo.cbSize = sizeof(m_menuInfo);
    m_menuInfo.fMask = MIIM_ID | MIIM_SUBMENU | MIIM_STRING | MIIM_STATE;

    //---------------------

    m_frameFont = NULL;
    m_titleFont = NULL;
    m_hdc = CreateCompatibleDC(NULL);
    m_iniFont = m_pCreateStyleFont((StyleItem*)m_pGetSettingPtr(SN_TOOLBAR));
    m_defFont = (HFONT)SelectObject(m_hdc, m_iniFont);

    //---------------------

    SendMessage(m_hCoreWnd = GetBBWnd(), BB_REGISTERMESSAGE, (WPARAM)m_hwnd, (LPARAM)bb_messages);
    Reconfig();

    //---------------------

    m_p_ID = ID_Item::ListStartup();    // start list of ID_Items
}

//---------------------------------------------------------------------------------

SystemMenuItem::~SystemMenuItem()
{
    DeleteObject((HFONT)SelectObject(m_hdc, m_defFont));
    DeleteDC(m_hdc);

    if (m_iniFont) DeleteObject(m_iniFont);
    if (m_frameFont) DeleteObject(m_frameFont);
    if (m_titleFont) DeleteObject(m_titleFont);

    DestroyWindow(m_hwnd);  // destroy window
    SendMessage(m_hCoreWnd, BB_UNREGISTERMESSAGE, (WPARAM)m_hwnd, (LPARAM)bb_messages);
    UnregisterClass(szClassName, m_hinstance);

    ID_Item::ListShutdown();            //  shutdown list of ID_Items
}

//---------------------------------------------------------------------------------

void SystemMenuItem::Reconfig()
{
    if (m_frameFont) DeleteObject(m_frameFont);
    if (m_titleFont) DeleteObject(m_titleFont);

    m_frameFont = m_pCreateStyleFont((StyleItem*)m_pGetSettingPtr(SN_MENUFRAME));
    m_titleFont = m_pCreateStyleFont((StyleItem*)m_pGetSettingPtr(SN_MENUTITLE));
}

//---------------------------------------------------------------------------------

LRESULT CALLBACK _SystemMenuProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message)
    {
        //--------------------------

        case _pSystemMenu->MSG_SYSTEM_MENU_SHOW:
            _pSystemMenu->ShowTaskMenu((HWND)lParam, true);
            break;

        //--------------------------

        case BB_BROADCAST:
            _pSystemMenu->ProcessBroam(lParam);
            break;

        //--------------------------

        case BB_RECONFIGURE:
            _pSystemMenu->Reconfig();
            break;

        //--------------------------

        default:
            return DefWindowProc(hwnd, message, wParam, lParam);
    }
    return 0;
}

//---------------------------------------------------------------------------------

void SystemMenuItem::ShowTaskMenu(HWND hwnd, bool popup)
{
    SIZE sz;
    bool bNOP;
    HMENU hMenu;
    TaskMenuLine *tm;
    int count, i = 0;
    int root_count = m_p_ID->SetCurrentIDCounter(popup);

    char tx[256];
    GetWindowText(hwnd, tx, sizeof(tx));

    SelectObject(m_hdc, m_frameFont);

    int pix_max = 16 + GetTaskMenuInfo(hwnd, &hMenu, &tm, &count, &bNOP);

    m_p_ID->XDelMenu(m_bbMenu);     // call delete on old root-menu

    int caption_len = strlen(tx);
    int *len_array = new int[caption_len];
    if (!caption_len)
        goto make_menu;

    SelectObject(m_hdc, m_titleFont);

    GetTextExtentExPoint(m_hdc, tx, caption_len, pix_max, &i, len_array, &sz);
    if (i < caption_len)
    {
        int *g = len_array + i;
        pix_max -= GetItemLength("...", &sz);
        while (*(--g) > pix_max) --i;
        char buf[200];
        strncpy(buf, tx, i);
        buf[i] =
            buf[i + 1] =
            buf[i + 2] = '.';
        buf[i + 3] = 0;
        m_bbMenu = m_p_ID->XMakeMenu(buf, popup);  // create menu object
    }
    else
make_menu:
        m_bbMenu = m_p_ID->XMakeMenu(tx, popup);  // create menu object

    m_p_ID->PushBack(hwnd); // add an ID_Item for this root-menu

    SelectObject(m_hdc, m_iniFont);  // select the toolbar font back in

    delete[] len_array;

    for (i = 0; i < count; ++i)
    {
        if (tm[i].m_buf[0])
            PickTaskMenuItem(hwnd, m_bbMenu, tm[i].m_hsubmenu, i, tm[i].m_state, tm[i].m_id, tm[i].m_buf, &bNOP, popup, root_count);
        else
            bNOP = true;
    }

    delete[] tm;
    ShowMenu(m_bbMenu);
}

//---------------------------------------------------------------------------------

int SystemMenuItem::GetItemLength(char *txt, SIZE *sz)
{
    GetTextExtentPoint32(m_hdc, txt, strlen(txt), sz);
    return sz->cx;
}

//---------------------------------------------------------------------------------

int SystemMenuItem::GetTaskMenuInfo(HWND hwnd, HMENU *p_hMenu, TaskMenuLine **tm2, int *p_count, bool *p_is_sbx)
{
    SIZE sz;
    BYTE b = 0;
    TaskMenuLine *tm;
    bool is_system = false;
    int len, max_len = 0, i = 0;

    SendMessageTimeout(hwnd, WM_INITMENU, (WPARAM)*p_hMenu, 0, SMTO_ABORTIFHUNG|SMTO_BLOCK, 100, NULL);

    *p_is_sbx = max_len > 0;
    *p_hMenu = GetSystemMenu(hwnd, false);
    *p_count = GetMenuItemCount(*p_hMenu);

    if (*p_count < 0)
    {
        GetSystemMenu(hwnd, true);                      // if the menu's empty, reset the default
        *p_hMenu = GetSystemMenu(hwnd, false);          // try again to get a handle
        *p_count = GetMenuItemCount(*p_hMenu);          // get the new menu-item count
    }

    *tm2 = new TaskMenuLine[*p_count];

    for (; i < *p_count; ++i)
    {
        tm = &(*tm2)[i];
        GetItemText(*p_hMenu, &len, i, tm->m_buf);

        ///////////////
        /*  // ....MFS_DISABLED = MFS_GRAYED; grayed items don't update this flag reliably
        if (m_menuInfo.fState & MFS_DISABLED)
            goto fill_tm;
        */
        ///////////////


        if (tm->m_buf[0])
        {
            CleanMenuItem(tm->m_buf, len);                      // clean up the menu-item
            max_len = max(max_len, GetItemLength(tm->m_buf, &sz));
        }
        else if (!is_system) continue;

        is_system = true;

//////////////
//fill_tm:
//////////////
        tm->m_hsubmenu = m_menuInfo.hSubMenu;
        tm->m_state = m_menuInfo.fState;
        tm->m_id = m_menuInfo.wID;
    }

    return max_len;
}

//---------------------------------------------------------------------------------

Menu *SystemMenuItem::MakeSystemSubmenu(HWND target_hwnd, char *name, HMENU hMenu, bool popup, int root_count)
{
    bool bNOP = false;
    char buf_local[TB_MAX];

    Menu *parent_menu = m_p_ID->XMakeMenu(name, popup);

    int len, i = 0, count = GetMenuItemCount(hMenu);

    if (!count)
        MakeMenuNOP(parent_menu, "[empty]");
    else do
    {
        GetItemText(hMenu, &len, i, buf_local);
        if (buf_local[0])
        {
            CleanMenuItem(buf_local, len);
            PickTaskMenuItem(target_hwnd, parent_menu, m_menuInfo.hSubMenu, i, m_menuInfo.fState, m_menuInfo.wID, buf_local, &bNOP, popup, root_count);
        }
        else
            bNOP = true;
    }
    while (++i < count);

    return parent_menu;
}

//---------------------------------------------------------------------------------

void SystemMenuItem::GetItemText(HMENU h_menu, int *p_len, int count, char *buffer)
{
    m_menuInfo.wID = NULL;
    m_menuInfo.dwTypeData = NULL;
    GetMenuItemInfo(h_menu, count, true, &m_menuInfo);  // get the length of the text
    *p_len = ++m_menuInfo.cch;
    m_menuInfo.dwTypeData = buffer;
    GetMenuItemInfo(h_menu, count, true, &m_menuInfo);  // get text into buffer and get id
    buffer[min(TB_MAX, *p_len)] = 0; // just use a big buffer and null-terminate it
}

//---------------------------------------------------------------------------------

void SystemMenuItem::PickTaskMenuItem(HWND target_hwnd, Menu *parent_menu, HMENU hsubmenu, int index, int state, int sys_command, char *buf, bool *nop, bool popup, int root_count)
{
    if (*nop)
        *nop = false, MakeMenuNOP(parent_menu, NULL);

    ///////////////////////////////////
    /* // ....MFS_DISABLED = MFS_GRAYED; grayed items don't update this flag reliably
    if (state & MFS_DISABLED)
        return;
    */
    ///////////////////////////////////

    if (hsubmenu)
    {
        SendMessageTimeout(target_hwnd, WM_INITMENUPOPUP, (WPARAM)hsubmenu, MAKELPARAM(index, TRUE), SMTO_ABORTIFHUNG|SMTO_BLOCK, 100, NULL);
        MakeSubmenu(parent_menu, MakeSystemSubmenu(target_hwnd, buf, hsubmenu, popup, root_count), buf);
    }
    else
    {
        sprintf(tm_buf_data, "@SystemMenuItem %d %d %d", target_hwnd, sys_command, root_count);  // build broam
        MakeMenuItem(parent_menu, buf, tm_buf_data, ((state & MFS_CHECKED) > 0));  // make menu-item
    }
}

//---------------------------------------------------------------------------------

void SystemMenuItem::CleanMenuItem(char *p, UINT len)
{
    bool is_tab = false;
    char *p2 = p + len - 2;

    while (1)
    {
        if (p2[0] == 9) is_tab = true; // peel off stuff aligned by tabs, like "ctrl+n"
        else if (is_tab && (p2[0] > 32)) { p2[1] = 0; break; }
        if (p2 == p) break;
        --p2;
    }

    do if (p[0] == '&') strcpy(p, p + 1);  // remove underlined letters for 0.90
    while ((++p)[0]);
}

//---------------------------------------------------------------------------------

void SystemMenuItem::ProcessBroam(LPARAM lParam)
{
    char *broam_temp = (char*)lParam;
    if (!_strnicmp(broam_temp, "@SystemMenuItem ", 16))
    {
        //-------------------
        HWND h;
        LPSTR tokens[3];
        int int_array[3];
        const int xob_max_line_length = 4096;
        char char_array[3][xob_max_line_length];
        tokens[0] = char_array[0];
        tokens[1] = char_array[1];
        tokens[2] = char_array[2];
        tokens[0][0] = tokens[1][0] = tokens[2][0] = 0;
        BBTokenize(broam_temp + 16, tokens, 3, NULL);
        int_array[0] = atoi(tokens[0]);
        int_array[1] = atoi(tokens[1]);
        int_array[2] = atoi(tokens[2]);
        //-------------------
        h = (HWND)int_array[0];
        PostMessage(h, WM_SYSCOMMAND, (WPARAM)int_array[1], 0);
        //-------------------
        m_p_ID->CleanList();            // clean out old ID_Items
        //-------------------
        UpdateMenu(h, int_array[2]);    // update menu
        //-------------------
        SetForegroundWindow(h);         // focus the window
    }
}

//---------------------------------------------------------------------------------

void SystemMenuItem::UpdateMenu(HWND hwnd, int id)
{
    if (m_p_ID->SetUpdateID(hwnd, id))
        ShowTaskMenu(hwnd, false);
}

//---------------------------------------------------------------------------------
